"""OpenAI Codex provider backed by agno OpenAILike."""

from asedclaw.providers.agno_provider import AgnoOpenAILikeProvider


class OpenAICodexProvider(AgnoOpenAILikeProvider):
    """Codex model route via agno OpenAILike wrapper."""


def _strip_model_prefix(model: str) -> str:
    """Compatibility helper retained for existing tests/callers."""
    if model.startswith("openai-codex/") or model.startswith("openai_codex/"):
        return model.split("/", 1)[1]
    return model
